from tensorflow.keras import layers
import tensorflow as tf
import numpy as np

'''
Base class for high order layers.
'''
class HighOrderLayer(layers.Layer):

    '''
    units is the number of units in the layer
    input_dim is the number of dimensions in the input
    '''
    def __init__(self, units=None, dtype=None):
        super(HighOrderLayer, self).__init__()
        self.numWeights = None
        self.units = None

    def build(self, input_shape):
        input_dim = int(input_shape[-1])

        w_init = tf.random_normal_initializer()
        self.w = tf.Variable(
            initial_value=w_init(
                shape=(
                    self.units,
                    input_dim,
                    self.numWeights),
                dtype=self.dtype),
            trainable=True)

        # We don't need b
        '''
        b_init = tf.zeros_initializer()
        self.b = tf.Variable(
            initial_value=b_init(
                shape=(
                    self.units,
                ),
                dtype=self.dtype),
            trainable=True)
        '''

    def function(self, x) :
        raise Exception('Function not implemented in '+type(self).__name__)
       
    def call(self, inputs):

        shapeIn = inputs.shape
        shape = self.w.shape
        res = self.function(inputs)
        res = tf.transpose(res,[1,2,0])
        res = tf.reshape(res, [-1, res.shape[1] * res.shape[2]])
        temp = tf.reshape(self.w, [-1, shape[1] * shape[2]])

        ans = tf.matmul(res, temp, transpose_a=False,
                        transpose_b=True)

        return ans



from .HighOrderLayer import *
import tensorflow as tf
import numpy as np
import math

'''
Fourier series layer takes the number of frequencies desired
The total number of weights is 2*frequencies+1.  The first component
is the offset, and then sin and cos for each frequency.
'''
class Fourier(HighOrderLayer):

    def __init__(self, units=None, frequencies=None, shift=0.0, length=2.0, dtype=None):
        '''
        Parameters :
            units - is the number of units in the layer
            frequencies - is the number of frequencies to include in fourier series
            shift - is an added phase shift
            length - is the wave length of the longest wave
            dtype - the data type for the tensorflow arrays
        '''
        super(Fourier, self).__init__(units=units,dtype=dtype)

        if units is None:
            print('You must define the units')
            raise

        if frequencies is None:
            print('You must define the frequencies.')
            raise

        self.units = units
        self.frequencies = frequencies
        self.numWeights = 2*frequencies+1
        self.length = length
        self.shift = shift

    def function(self, inputs) :
        return self.basisFourier(inputs, self.frequencies)

    def basisFourier(self, x, numFrequencies) :
        series = tf.convert_to_tensor([0*x+0.5-self.shift])
        for i in range(1,1+numFrequencies) :
            other = tf.convert_to_tensor([tf.math.cos(math.pi*i*x/self.length),tf.math.sin(math.pi*i*x/self.length)])
            series = tf.concat([series,other],axis=0)
        
        return series



from . HighOrderLayer import *
from . Functions import *
import tensorflow as tf
import numpy as np

"""
Store the function and the number of weights in the
same object.
"""
class FunctionWrapper():
    def __init__(self, basis, numWeights):
        self.numWeights = numWeights
        self.basis = basis

    def __call__(self, val):
        return self.basis(val)

b0 = FunctionWrapper(basis0, 1)

b1 = FunctionWrapper(basis1, 2)
b1D = FunctionWrapper(basis1DG, 4)
b1C = FunctionWrapper(basis1CG, 3)

b2 = FunctionWrapper(basis2, 3)
b2C = FunctionWrapper(basis2CG, 5)
b2D = FunctionWrapper(basis2DG, 6)

b3 = FunctionWrapper(basis3, 4)
b3C = FunctionWrapper(basis3CG, 7)
b3D = FunctionWrapper(basis3DG, 8)

b4 = FunctionWrapper(basis4, 5)
b4C = FunctionWrapper(basis4CG, 9)
b4D = FunctionWrapper(basis4DG, 10)

b5 = FunctionWrapper(basis5, 6)
b5C = FunctionWrapper(basis5CG, 11)
b5D = FunctionWrapper(basis5DG, 12)

'''
Piecewise Lagrange polynomials with gauss lobatto points for tensorflow!
TODO: might move the polynomial derivation inside so the user does not need
to specify a basis, only the order and number of knots.
'''
class Polynomial(HighOrderLayer):
    def __init__(self, units=None, basis=None, shift=0.0,dtype=None):
        '''
        Parameters:
            units - is the number of units in the layer.
            basis - is an instance of the "Function" class which contains a basis and the number of weights.
            shift - shift the polynomial center by the given value.
            dtype - the data type for the weight arrays
        '''
        super(Polynomial, self).__init__(units=units,dtype=dtype)

        if units is None:
            print('You must define the units')
            raise

        if basis is None:
            print('You must define the basis.')
            raise

        self.units = units
        self.basis = basis
        self.shift = shift
        self.numWeights = self.basis.numWeights

    def function(self, inputs) :
        return self.basis(inputs-self.shift)



from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Layer
import tensorflow as tf
import copy


#Just evaluate the basis function on the input and reshape to the desired form
class ExpansionLayer2D(Layer) :
    def __init__(self, basis=None) :
        super().__init__()
        if basis==None :
            raise Exception('You must define the basis function in ExpansionLayer2D')
        self.basis = basis

    def build(self,input_shape) :
        pass
    
    def call(self, inputs) :
        res = self.basis(inputs)
        res = tf.transpose(res, [1, 2, 3, 4, 0])
        ts = res.get_shape()
        res = tf.reshape(res, [-1,res.shape[1],res.shape[2],res.shape[3]*res.shape[4]])
        return res

def high_order_convolution2D(
            x,
            filters,
            kernel_size,
            basis=None,
            strides=(1, 1),
            padding='valid',
            data_format=None,
            dilation_rate=(1, 1),
            activation=None,
            use_bias=True,
            kernel_initializer='glorot_uniform',
            bias_initializer='zeros',
            kernel_regularizer=None,
            bias_regularizer=None,
            activity_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
            **kwargs) :

    x=ExpansionLayer2D(basis=basis)(x)
    x = Conv2D(
        filters,
        kernel_size,
        strides=strides,
        padding=padding,
        data_format=data_format,
        dilation_rate=dilation_rate,
        activation=None,
        use_bias=use_bias,
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        bias_regularizer=bias_regularizer,
        activity_regularizer=activity_regularizer,
        kernel_constraint=kernel_constraint,
        bias_constraint=bias_constraint,
        **kwargs)(x)

    return x





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="high_order_layers",
    version="1.1.7",
    author="John Loverich",
    author_email="john.loverich@gmail.com",
    description="Polynomial, piecewise polynomial, fourier series layers for tensorflow",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jloveric/high-order-layers",
    packages=setuptools.find_packages(),
    install_requires=['numpy','tensorflow'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
) 



import tensorflow as tf
import numpy as np
import sys
import math

# Polynomial basis functions
def basis0(x):
    return tf.convert_to_tensor([x])

def basis1(x):
    return tf.convert_to_tensor([-0.5 * (x - 1.0), 0.5 * (x + 1.0)])

# quadratic polynomial
def basis2(xIn):
    x = tf.convert_to_tensor(xIn)
    return tf.convert_to_tensor(
        [0.5 * x * (x - 1), -1.0 * (x + 1) * (x - 1), 0.5 * x * (x + 1)])

# cubic polynomial
def basis3(xIn):
    x = xIn
    eta = x
    powEta2 = eta * x
    powEta3 = powEta2 * x

    temp = tf.convert_to_tensor([-0.6666666666666666 * powEta3 + 0.6666666666666666 * powEta2 +
                                 0.1666666666666667 * eta - 0.1666666666666667,
                                 1.333333333333333 * powEta3 - 0.6666666666666666 * powEta2 -
                                 1.333333333333333 * eta + 0.6666666666666666,
                                 -1.333333333333333 * powEta3 - 0.6666666666666666 * powEta2 +
                                 1.333333333333333 * eta + 0.6666666666666666,
                                 0.6666666666666666 * powEta3 + 0.6666666666666666 * powEta2 -
                                 0.1666666666666667 * eta - 0.1666666666666667])
    return temp

# quartic polynomial
def basis4(xIn):
    x = xIn

    eta = x
    powEta2 = x * x
    powEta3 = powEta2 * x
    powEta4 = powEta3 * x

    temp = tf.convert_to_tensor([powEta4 -
                                 powEta3 -
                                 0.5 *
                                 powEta2 +
                                 0.5 *
                                 eta, -
                                 2.0 *
                                 powEta4 +
                                 1.414213562373095 *
                                 powEta3 +
                                 2.0 *
                                 powEta2 -
                                 1.414213562373095 *
                                 eta, 1 -
                                 3 *
                                 powEta2 +
                                 2 *
                                 powEta4, -
                                 2 *
                                 powEta4 -
                                 1.414213562373095 *
                                 powEta3 +
                                 2.0 *
                                 powEta2 +
                                 1.414213562373095 *
                                 eta, powEta4 +
                                 powEta3 -
                                 0.5 *
                                 powEta2 -
                                 0.5 *
                                 eta])

    return temp

# quintic polynomial


def basis5(xIn):
    #x = np.tanh(xIn)
    x = xIn
    eta = x
    powEta2 = eta * eta
    powEta3 = eta * powEta2
    powEta4 = eta * powEta3
    powEta5 = eta * powEta4

    temp = tf.convert_to_tensor([(0.1 -
                                  0.1 *
                                  eta -
                                  1.2 *
                                  powEta2 +
                                  1.2 *
                                  powEta3 +
                                  1.6 *
                                  powEta4 -
                                  1.6 *
                                  powEta5), 3.2 *
                                 powEta5 -
                                 2.588854381999832 *
                                 powEta4 -
                                 3.505572809000084 *
                                 powEta3 +
                                 2.83606797749979 *
                                 powEta2 +
                                 0.3055728090000842 *
                                 eta -
                                 0.247213595499958, -
                                 3.2 *
                                 powEta5 +
                                 0.9888543819998319 *
                                 powEta4 +
                                 5.294427190999916 *
                                 powEta3 -
                                 1.63606797749979 *
                                 powEta2 -
                                 2.094427190999916 *
                                 eta +
                                 0.6472135954999582, 3.2 *
                                 powEta5 +
                                 0.9888543819998319 *
                                 powEta4 -
                                 5.294427190999916 *
                                 powEta3 -
                                 1.63606797749979 *
                                 powEta2 +
                                 2.094427190999916 *
                                 eta +
                                 0.6472135954999582, -
                                 3.2 *
                                 powEta5 -
                                 2.588854381999832 *
                                 powEta4 +
                                 3.505572809000084 *
                                 powEta3 +
                                 2.83606797749979 *
                                 powEta2 -
                                 0.3055728090000842 *
                                 eta -
                                 0.247213595499958, (0.1 +
                                                     0.1 *
                                                     eta -
                                                     1.2 *
                                                     powEta2 -
                                                     1.2 *
                                                     powEta3 +
                                                     1.6 *
                                                     powEta4 +
                                                     1.6 *
                                                     powEta5)])

    return temp

'''
Discontinuous polynomial, 2 parts
'''
def basisDG(x, basis) :
    xr = tf.where(x > 0, 2.0 * (x - 0.5), 0 * x)
    xl = tf.where(x <= 0, 2.0 * (x + 0.5), 0 * x)
    pos = tf.where(x > 0, 1.0, 0.0)
    neg = tf.where(x <= 0, 1.0, 0.0)

    fr = pos*basis(xr)
    fl = neg*basis(xl)
    f = tf.concat([fl,fr],axis=0)
    
    return f

def basis1DG(x):
    return basisDG(x, basis1)

def basis2DG(x):
    return basisDG(x, basis2)

def basis3DG(x):
    return basisDG(x, basis3)

def basis4DG(x):
    return basisDG(x, basis4)

def basis5DG(x):
    return basisDG(x, basis5)

'''
Continuous piecewise polynomials
'''
def basisCG(x, basis, width) :
    #Add one additional dimension for the batch, I believe.
    dims = len(x.get_shape().as_list())+1
    
    lR = [ [0]*2 for _ in range(dims) ]
    lL = [ [0]*2 for _ in range(dims) ]

    lR[0][0] = width-1
    lL[0][1] = width-1

    paddingsR = tf.convert_to_tensor(lR)
    paddingsL = tf.convert_to_tensor(lL)

    xr = tf.where(x > 0, 2.0 * (x - 0.5), 0 * x)
    xl = tf.where(x <= 0, 2.0 * (x + 0.5), 0 * x)
    pos = tf.where(x > 0, 1.0, 0.0)
    neg = tf.where(x <= 0, 1.0, 0.0)

    fr = tf.pad(pos*basis(xr),paddingsR)
    fl = tf.pad(neg*basis(xl),paddingsL)
    
    f = fr+fl
    
    return f

def basis1CG(x):
    return basisCG(x, basis1, 2)

def basis2CG(x):
    return basisCG(x, basis2, 3)

def basis3CG(x):
    return basisCG(x, basis3, 4)

def basis4CG(x):
    return basisCG(x, basis4, 5)

def basis5CG(x):
    return basisCG(x, basis5, 6)
    
